import io
import pytest
import re
import yamale

from . import get_fixture
from .. import validators as val

types = {"schema": "types.yaml", "bad": "types_bad_data.yaml", "good": "types_good_data.yaml"}

nested = {"schema": "nested.yaml", "bad": "nested_bad_data.yaml", "good": "nested_good_data.yaml"}

custom = {"schema": "custom_types.yaml", "bad": "custom_types_bad.yaml", "good": "custom_types_good.yaml"}

keywords = {"schema": "keywords.yaml", "bad": "keywords_bad.yaml", "good": "keywords_good.yaml"}

lists = {"schema": "lists.yaml", "bad": "lists_bad.yaml", "bad2": "lists_bad2.yaml", "good": "lists_good.yaml"}

maps = {"schema": "map.yaml", "bad": "map_bad.yaml", "bad2": "map_bad2.yaml", "good": "map_good.yaml"}

anys = {"schema": "any.yaml", "bad": "any_bad.yaml", "good": "any_good.yaml"}

any_undefined = {"schema": "any_undefined_schema.yaml", "bad": "any_undefined.yaml"}

list_include = {"schema": "list_include.yaml", "good": "list_include_good.yaml"}

issue_22 = {"schema": "issue_22.yaml", "good": "issue_22_good.yaml"}

issue_50 = {"schema": "issue_50.yaml", "good": "issue_50_good.yaml"}

regexes = {"schema": "regex.yaml", "bad": "regex_bad.yaml", "good": "regex_good.yaml"}

ips = {"schema": "ip.yaml", "bad": "ip_bad.yaml", "good": "ip_good.yaml"}

macs = {"schema": "mac.yaml", "bad": "mac_bad.yaml", "good": "mac_good.yaml"}

nested_map = {"schema": "nested_map.yaml", "good": "nested_map_good.yaml"}

top_level_map = {"schema": "top_level_map.yaml", "good": "top_level_map_good.yaml"}

semver = {"schema": "semver_schema.yaml", "good": "semver_good.yaml", "bad": "semver_bad.yaml"}


include_validator = {
    "schema": "include_validator.yaml",
    "good": "include_validator_good.yaml",
    "bad": "include_validator_bad.yaml",
}

strict_map = {"schema": "strict_map.yaml", "good": "strict_map_good.yaml", "bad": "strict_map_bad.yaml"}

mixed_strict_map = {
    "schema": "mixed_strict_map.yaml",
    "good": "mixed_strict_map_good.yaml",
    "bad": "mixed_strict_map_bad.yaml",
}

strict_list = {"schema": "strict_list.yaml", "good": "strict_list_good.yaml", "bad": "strict_list_bad.yaml"}

nested_map2 = {"schema": "nested_map2.yaml", "good": "nested_map2_good.yaml", "bad": "nested_map2_bad.yaml"}

static_list = {"schema": "static_list.yaml", "good": "static_list_good.yaml", "bad": "static_list_bad.yaml"}

nested_issue_54 = {"schema": "nested.yaml", "bad": "nested_issue_54.yaml", "good": "nested_good_data.yaml"}

map_key_constraint = {
    "schema": "map_key_constraint.yaml",
    "good": "map_key_constraint_good.yaml",
    "bad_base": "map_key_constraint_bad_base.yaml",
    "bad_nest": "map_key_constraint_bad_nest.yaml",
    "bad_nest_con": "map_key_constraint_bad_nest_con.yaml",
}

numeric_bool_coercion = {
    "schema": "numeric_bool_coercion.yaml",
    "good": "numeric_bool_coercion_good.yaml",
    "bad": "numeric_bool_coercion_bad.yaml",
}

subset = {
    "schema": "subset.yaml",
    "good": "subset_good.yaml",
    "good2": "subset_good2.yaml",
    "bad": "subset_bad.yaml",
    "bad2": "subset_bad2.yaml",
    "bad3": "subset_bad3.yaml",
}

subset_empty = {"schema": "subset_empty.yaml", "good": "subset_empty_good.yaml", "good2": "subset_empty_good2.yaml"}

subset_nodef = {"schema": "subset_nodef.yaml"}

test_data = [
    types,
    nested,
    custom,
    keywords,
    lists,
    maps,
    anys,
    any_undefined,
    list_include,
    issue_22,
    issue_50,
    regexes,
    ips,
    macs,
    nested_map,
    top_level_map,
    include_validator,
    strict_map,
    mixed_strict_map,
    strict_list,
    nested_map2,
    static_list,
    nested_issue_54,
    map_key_constraint,
    numeric_bool_coercion,
    semver,
    subset,
    subset_empty,
]

for d in test_data:
    for key in d.keys():
        if key == "schema":
            d[key] = yamale.make_schema(get_fixture(d[key]))
        else:
            d[key] = yamale.make_data(get_fixture(d[key]))


def test_flat_make_schema():
    assert isinstance(types["schema"]._schema["string"], val.String)


def test_nested_schema():
    nested_schema = nested["schema"]._schema
    assert isinstance(nested_schema["string"], val.String)
    assert isinstance(nested_schema["list"], (list, tuple))
    assert isinstance(nested_schema["list"][0], val.String)


@pytest.mark.parametrize("data_map", test_data)
def test_good(data_map):
    for k, good_data in data_map.items():
        if k.startswith("good"):
            yamale.validate(data_map["schema"], good_data)


def test_bad_validate():
    assert count_exception_lines(types["schema"], types["bad"]) == 9


def test_bad_nested():
    assert count_exception_lines(nested["schema"], nested["bad"]) == 2


def test_bad_nested_issue_54():
    exp = [
        "string: Required field missing",
        "number: Required field missing",
        "integer: Required field missing",
        "boolean: Required field missing",
        "date: Required field missing",
        "datetime: Required field missing",
        "nest: Required field missing",
        "list: Required field missing",
    ]
    match_exception_lines(nested_issue_54["schema"], nested_issue_54["bad"], exp)


def test_bad_custom():
    assert count_exception_lines(custom["schema"], custom["bad"]) == 1


def test_bad_lists():
    assert count_exception_lines(lists["schema"], lists["bad"]) == 6


def test_bad2_lists():
    assert count_exception_lines(lists["schema"], lists["bad2"]) == 2


def test_bad_maps():
    assert count_exception_lines(maps["schema"], maps["bad"]) == 7


def test_bad_maps2():
    assert count_exception_lines(maps["schema"], maps["bad2"]) == 1


def test_bad_keywords():
    assert count_exception_lines(keywords["schema"], keywords["bad"]) == 9


def test_bad_anys():
    assert count_exception_lines(anys["schema"], anys["bad"]) == 5


def test_undefined_include():
    assert count_exception_lines(any_undefined["schema"], any_undefined["bad"]) == 1


def test_bad_semver():
    assert count_exception_lines(semver["schema"], semver["bad"]) == 1


def test_bad_regexes():
    assert count_exception_lines(regexes["schema"], regexes["bad"]) == 4


def test_bad_include_validator():
    exp = ["key1: 'a_string' is not a int."]
    match_exception_lines(include_validator["schema"], include_validator["bad"], exp)


def test_bad_schema():
    with pytest.raises(SyntaxError) as excinfo:
        yamale.make_schema(get_fixture("bad_schema.yaml"))
    assert "fixtures/bad_schema.yaml" in str(excinfo.value)


def test_empty_schema():
    with pytest.raises(ValueError) as excinfo:
        yamale.make_schema(get_fixture("empty_schema.yaml"))
    assert "empty_schema.yaml is an empty file!" in str(excinfo.value)


@pytest.mark.parametrize(
    "schema_filename", ["bad_schema_rce.yaml", "bad_schema_rce2.yaml", "bad_schema_rce3.yaml", "bad_schema_rce4.yaml"]
)
def test_vulnerable_schema(schema_filename):
    with pytest.raises(SyntaxError) as excinfo:
        yamale.make_schema(get_fixture(schema_filename))
    assert schema_filename in str(excinfo.value)


def test_list_is_not_a_map():
    exp = [" : '[1, 2]' is not a map"]
    match_exception_lines(strict_map["schema"], strict_list["good"], exp)


def test_bad_strict_map():
    exp = ["extra: Unexpected element"]
    match_exception_lines(strict_map["schema"], strict_map["bad"], exp, strict=True)


def test_bad_strict_list():
    exp = ["2: Unexpected element"]
    match_exception_lines(strict_list["schema"], strict_list["bad"], exp, strict=True)


def test_bad_mixed_strict_map():
    exp = ["field3.extra: Unexpected element"]
    match_exception_lines(mixed_strict_map["schema"], mixed_strict_map["bad"], exp)


def test_bad_nested_map2():
    exp = ["field1.field1_1: Required field missing"]
    match_exception_lines(nested_map2["schema"], nested_map2["bad"], exp)


def test_bad_static_list():
    exp = ["0: Required field missing"]
    match_exception_lines(static_list["schema"], static_list["bad"], exp)


def test_bad_map_key_constraint_base():
    exp = [": Key error - 'bad' is not a int."]
    match_exception_lines(map_key_constraint["schema"], map_key_constraint["bad_base"], exp)


def test_bad_map_key_constraint_nest():
    exp = ["1.0: Key error - '100' is not a str."]
    match_exception_lines(map_key_constraint["schema"], map_key_constraint["bad_nest"], exp)


def test_bad_map_key_constraint_nest_con():
    exp = [
        "1.0: Key error - '100' is not a str.",
        "1.0: Key error - 'baz' contains excluded character 'z'",
    ]
    match_exception_lines(map_key_constraint["schema"], map_key_constraint["bad_nest_con"], exp)


def test_bad_numeric_bool_coercion():
    exp = [
        "integers.0: 'False' is not a int.",
        "integers.1: 'True' is not a int.",
        "numbers.0: 'False' is not a num.",
        "numbers.1: 'True' is not a num.",
    ]
    match_exception_lines(numeric_bool_coercion["schema"], numeric_bool_coercion["bad"], exp)


def test_bad_subset():
    exp = ["subset_list: 'subset' may not be an empty set."]
    match_exception_lines(subset["schema"], subset["bad"], exp)


def test_bad_subset2():
    exp = ["subset_list: '[1]' is not a int.", "subset_list: '[1]' is not a str."]
    match_exception_lines(subset["schema"], subset["bad2"], exp)


def test_bad_subset3():
    exp = ["subset_list: '{'a': 1}' is not a int.", "subset_list: '{'a': 1}' is not a str."]
    match_exception_lines(subset["schema"], subset["bad3"], exp)


def test_nodef_subset_schema():
    with pytest.raises(ValueError) as e:
        yamale.make_schema(get_fixture(subset_nodef["schema"]))

    assert "'subset' requires at least one validator!" in str(e.value)


@pytest.mark.parametrize(
    "use_schema_string,use_data_string,expected_message_re",
    [
        (False, False, "^Error validating data '.*?' with schema '.*?'\n\t"),
        (True, False, "^Error validating data '.*?'\n\t"),
        (False, True, "^Error validating data with schema '.*?'\n\t"),
        (True, True, "^Error validating data\n\t"),
    ],
)
def test_validate_errors(use_schema_string, use_data_string, expected_message_re):
    schema_path = get_fixture("types.yaml")
    data_path = get_fixture("types_bad_data.yaml")
    if use_schema_string:
        with io.open(schema_path, encoding="utf-8") as f:
            schema = yamale.make_schema(content=f.read())
    else:
        schema = yamale.make_schema(schema_path)
    if use_data_string:
        with io.open(data_path, encoding="utf-8") as f:
            data = yamale.make_data(content=f.read())
    else:
        data = yamale.make_data(data_path)
    with pytest.raises(yamale.yamale_error.YamaleError) as excinfo:
        yamale.validate(schema, data)
    assert re.match(expected_message_re, excinfo.value.message, re.MULTILINE), "Message {} should match {}".format(
        excinfo.value.message, expected_message_re
    )


def match_exception_lines(schema, data, expected, strict=False):
    with pytest.raises(ValueError) as e:
        yamale.validate(schema, data, strict)

    got = e.value.results[0].errors
    got.sort()
    expected.sort()
    assert got == expected


def count_exception_lines(schema, data, strict=False):
    with pytest.raises(ValueError) as e:
        yamale.validate(schema, data, strict)
    result = e.value.results[0]
    return len(result.errors)



import io
import pytest
from .. import yaml_reader
from yamale.tests import get_fixture

parsers = ["pyyaml", "PyYAML", "ruamel"]
TYPES = get_fixture("types.yaml")
NESTED = get_fixture("nested.yaml")
KEYWORDS = get_fixture("keywords.yaml")


@pytest.mark.parametrize("parser", parsers)
@pytest.mark.parametrize("use_string", [True, False])
def test_parse(parser, use_string):
    if use_string:
        with io.open(TYPES, encoding="utf-8") as f:
            content = f.read()
        a = yaml_reader.parse_yaml(parser=parser, content=content)[0]
    else:
        a = yaml_reader.parse_yaml(TYPES, parser)[0]
    assert a["string"] == "str()"


def test_parse_validates_arguments():
    with pytest.raises(TypeError):
        yaml_reader.parse_yaml(path=TYPES, content="name: Bob")
    with pytest.raises(TypeError):
        yaml_reader.parse_yaml(path=None, content=None)


@pytest.mark.parametrize("parser", parsers)
def test_types(parser):
    t = yaml_reader.parse_yaml(TYPES, parser)[0]
    assert t["string"] == "str()"
    assert t["number"] == "num()"
    assert t["boolean"] == "bool()"
    assert t["integer"] == "int()"


@pytest.mark.parametrize("parser", parsers)
def test_keywords(parser):
    t = yaml_reader.parse_yaml(KEYWORDS, parser)[0]
    assert t["optional_min"] == "int(min=1, required=False)"


@pytest.mark.parametrize("parser", parsers)
def test_nested(parser):
    t = yaml_reader.parse_yaml(NESTED, parser)[0]
    assert t["list"][-1]["string"] == "str()"



class Validator(object):
    """Base class for all Validators"""

    constraints = []
    value_type = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        # Is field required? Default is True
        self.is_required = bool(kwargs.pop("required", True))

        # Can value be None if field is optional? Default is True
        self._value_can_be_none = bool(kwargs.pop("none", True))

        # Construct all constraints
        self._constraints_inst = self._create_constraints(self.constraints, self.value_type, kwargs)

    def _create_constraints(self, constraint_classes, value_type, kwargs):
        constraints = []
        for constraint in constraint_classes:
            constraints.append(constraint(value_type, kwargs))
        return constraints

    @property
    def tag(self):
        return self.__class__

    @property
    def is_optional(self):
        return not self.is_required

    @property
    def can_be_none(self):
        """Check if value for optional field can be None."""
        return self._value_can_be_none

    def _is_valid(self, value):
        """Validators must implement this. Return True if value is valid."""
        raise NotImplementedError("You need to override this function")

    def get_name(self):
        return self.tag

    def validate(self, value):
        """
        Check if ``value`` is valid.

        :returns: [errors] If ``value`` is invalid, otherwise [].
        """
        errors = []

        # Make sure the type validates first.
        valid = self._is_valid(value)
        if not valid:
            errors.append(self.fail(value))
            return errors

        # Then validate all the constraints second.
        for constraint in self._constraints_inst:
            error = constraint.is_valid(value)
            if error:
                if isinstance(error, list):
                    errors.extend(error)
                else:
                    errors.append(error)

        return errors

    def is_valid(self, value):
        return self.validate(value) == []

    def fail(self, value):
        """Override to define a custom fail message"""
        return "'%s' is not a %s." % (value, self.get_name())

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.args, self.kwargs)

    def __eq__(self, other):
        # Validators are equal if they have the same args and kwargs.
        eq = [isinstance(other, self.__class__), self.args == other.args, self.kwargs == other.kwargs]
        return all(eq)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Validate yaml files and check them against their schemas. Designed to be used outside of Vagrant.

Just install Yamale:
    pip install yamale
"""

import argparse
import glob
import os
import re
import multiprocessing
from .yamale_error import YamaleError
from .schema.validationresults import Result
from .version import __version__

import yamale

schemas = {}


def _validate(schema_path, data_path, parser, strict, _raise_error):
    schema = schemas.get(schema_path)
    try:
        if not schema:
            schema = yamale.make_schema(schema_path, parser)
            schemas[schema_path] = schema
    except (SyntaxError, ValueError) as e:
        results = [Result([str(e)])]
        if not _raise_error:
            return results
        raise YamaleError(results)
    data = yamale.make_data(data_path, parser)
    return yamale.validate(schema, data, strict, _raise_error)


def _find_data_path_schema(data_path, schema_name):
    """Starts in the data file folder and recursively looks
    in parents for `schema_name`"""
    if not data_path or data_path == os.path.abspath(os.sep) or data_path == ".":
        return None
    directory = os.path.dirname(data_path)
    path = glob.glob(os.path.join(directory, schema_name))
    if not path:
        return _find_schema(directory, schema_name)
    return path[0]


def _find_schema(data_path, schema_name):
    """Checks if `schema_name` is a valid file, if not
    searches in `data_path` for it."""

    if os.path.isfile(schema_name):
        return schema_name

    directory = os.path.dirname(data_path)
    path = glob.glob(os.path.join(directory, schema_name))
    for p in path:
        if os.path.isfile(p):
            return p

    return _find_data_path_schema(data_path, schema_name)


def _validate_file(yaml_path, schema_name, parser, strict, should_exclude):
    if should_exclude(yaml_path):
        return
    s = _find_schema(yaml_path, schema_name)
    if not s:
        raise ValueError("Invalid schema name for '{}' or schema not found.".format(schema_name))
    _validate(s, yaml_path, parser, strict, True)


def _validate_dir(root, schema_name, cpus, parser, strict, should_exclude):
    pool = multiprocessing.Pool(processes=cpus)
    res = []
    error_messages = []
    for root, _, files in os.walk(root):
        for f in files:
            if (f.endswith(".yaml") or f.endswith(".yml")) and f != schema_name:
                yaml_path = os.path.join(root, f)
                if should_exclude(yaml_path):
                    continue
                schema_path = _find_schema(yaml_path, schema_name)
                if schema_path:
                    res.append(pool.apply_async(_validate, (schema_path, yaml_path, parser, strict, False)))
                else:
                    print(f"No schema found for: {yaml_path}")

    print(f"Found {len(res)} yaml files to validate...")
    for r in res:
        sub_results = r.get(timeout=300)
        error_messages.extend([str(sub_result) for sub_result in sub_results if not sub_result.isValid()])
    pool.close()
    pool.join()
    if error_messages:
        raise ValueError("\n----\n".join(set(error_messages)))


def _router(paths, schema_name, cpus, parser, excludes=None, strict=True, verbose=False):
    EXCLUDE_REGEXES = tuple(re.compile(e) for e in excludes) if excludes else tuple()

    def should_exclude(yaml_path):
        has_match = any(pattern.search(yaml_path) for pattern in EXCLUDE_REGEXES)
        if has_match and verbose:
            print("Skipping validation for %s due to exclude pattern" % yaml_path)
        return has_match

    for path in paths:
        abs_path = os.path.abspath(path)
        if os.path.exists(abs_path):
            print(f"Validating {path}...")
        else:
            raise ValueError(f"Path does not exist: {path}")

        if os.path.isdir(abs_path):
            _validate_dir(abs_path, schema_name, cpus, parser, strict, should_exclude)
        else:
            _validate_file(abs_path, schema_name, parser, strict, should_exclude)


def main():
    def int_or_auto(num_cpu):
        if num_cpu == "auto":
            return multiprocessing.cpu_count()
        return int(num_cpu)

    parser = argparse.ArgumentParser(description="Validate yaml files.")
    parser.add_argument(
        "paths",
        metavar="PATH",
        default=["./"],
        nargs="*",
        help="Paths to validate, either directories or files. Default is the current directory.",
    )
    parser.add_argument("-s", "--schema", default="schema.yaml", help="filename of schema. Default is schema.yaml.")
    parser.add_argument(
        "-e",
        "--exclude",
        metavar="PATTERN",
        action="append",
        help="Python regex used to exclude files from validation. Any substring match of a file's absolute path will be excluded. Uses default Python3 regex. Option can be supplied multiple times.",
    )
    parser.add_argument(
        "-p",
        "--parser",
        default="pyyaml",
        help='YAML library to load files. Choices are "ruamel" or "pyyaml" (default).',
    )
    parser.add_argument(
        "-n",
        "--cpu-num",
        default=4,
        type=int_or_auto,
        help="Number of child processes to spawn for validation. Default is 4. 'auto' to use CPU count.",
    )
    parser.add_argument(
        "-x",
        "--no-strict",
        action="store_true",
        help="Disable strict mode, unexpected elements in the data will be accepted.",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="show verbose information")
    parser.add_argument("-V", "--version", action="version", version=__version__)
    args = parser.parse_args()
    try:
        _router(
            paths=args.paths,
            schema_name=args.schema,
            cpus=args.cpu_num,
            parser=args.parser,
            excludes=args.exclude,
            strict=not args.no_strict,
            verbose=args.verbose,
        )
    except (SyntaxError, NameError, TypeError, ValueError) as e:
        print("Validation failed!\n%s" % str(e))
        exit(1)
    try:
        print("Validation success! 👍")
    except UnicodeEncodeError:
        print("Validation success!")


if __name__ == "__main__":
    main()



from datetime import date, datetime
from yamale import validators as val


def test_validator_defaults():
    """
    Unit test the dictionary of default validators.
    """
    assert val.DefaultValidators[val.String.tag] is val.String
    assert val.DefaultValidators[val.Any.__name__] is val.Any


def test_equality():
    assert val.String() == val.String()
    assert val.String(hello="wat") == val.String(hello="wat")
    assert val.String(hello="wat") != val.String(hello="nope")
    assert val.Boolean("yep") != val.Boolean("nope")


def test_integer():
    v = val.Integer()
    assert v.is_valid(1)
    assert not v.is_valid("1")
    assert not v.is_valid(1.34)


def test_string():
    v = val.String()
    assert v.is_valid("1")
    assert not v.is_valid(1)


def test_regex():
    v = val.Regex(r"^(abc)\1?de$", name="test regex")
    assert v.is_valid("abcabcde")
    assert not v.is_valid("abcabcabcde")
    assert not v.is_valid("\12")
    assert v.fail("woopz") == "'woopz' is not a test regex."

    v = val.Regex(r"[a-z0-9]{3,}s\s$", ignore_case=True)
    assert v.is_valid("b33S\v")
    assert v.is_valid("B33s\t")
    assert not v.is_valid(" b33s ")
    assert not v.is_valid("b33s  ")
    assert v.fail("fdsa") == "'fdsa' is not a regex match."

    v = val.Regex(r"A.+\d$", ignore_case=False, multiline=True)
    assert v.is_valid("A_-3\n\n")
    assert not v.is_valid("a!!!!!5\n\n")

    v = val.Regex(r".*^Ye.*s\.", ignore_case=True, multiline=True, dotall=True)
    assert v.is_valid("YEeeEEEEeeeeS.")
    assert v.is_valid("What?\nYes!\nBEES.\nOK.")
    assert not v.is_valid("YES-TA-TOES?")
    assert not v.is_valid("\n\nYaes.")


def test_number():
    v = val.Number()
    assert v.is_valid(1)
    assert v.is_valid(1.3425235)
    assert not v.is_valid("str")


def test_boolean():
    v = val.Boolean()
    assert v.is_valid(True)
    assert v.is_valid(False)
    assert not v.is_valid("")
    assert not v.is_valid(0)


def test_date():
    v = val.Day()
    assert v.is_valid(date(2015, 1, 1))
    assert v.is_valid(datetime(2015, 1, 1, 1))
    assert not v.is_valid("")
    assert not v.is_valid(0)


def test_datetime():
    v = val.Timestamp()
    assert v.is_valid(datetime(2015, 1, 1, 1))
    assert not v.is_valid(date(2015, 1, 1))
    assert not v.is_valid("")
    assert not v.is_valid(0)


def test_list():
    v = val.List()
    assert v.is_valid([])
    assert v.is_valid(())
    assert not v.is_valid("")
    assert not v.is_valid(0)


def test_null():
    v = val.Null()
    assert v.is_valid(None)
    assert not v.is_valid("None")
    assert not v.is_valid(0)
    assert not v.is_valid(float("nan"))
    assert not v.is_valid({})


def test_ip():
    v = val.Ip()
    assert v.is_valid("192.168.1.1")
    assert v.is_valid("192.168.1.255")
    assert v.is_valid("192.168.3.1/24")
    assert v.is_valid("2001:db8::")
    assert v.is_valid("2001:db8::/64")
    assert not v.is_valid("257.192.168.1")
    assert not v.is_valid("192.168.1.256")
    assert not v.is_valid("2001:db8::/129")
    assert not v.is_valid("2001:dg8::/127")
    assert not v.is_valid("asdf")


def test_mac():
    v = val.Mac()
    assert v.is_valid("12:34:56:78:90:ab")
    assert v.is_valid("1234:5678:90ab")
    assert v.is_valid("12-34-56-78-90-ab")
    assert v.is_valid("1234-5678-90ab")

    assert v.is_valid("12:34:56:78:90:AB")
    assert v.is_valid("1234:5678:90AB")
    assert v.is_valid("12-34-56-78-90-AB")
    assert v.is_valid("1234-5678-90AB")

    assert v.is_valid("ab:cd:ef:12:34:56")
    assert v.is_valid("abcd:ef12:3456")
    assert v.is_valid("ab-cd-ef-12-34-56")
    assert v.is_valid("abcd-ef12-3456")

    assert v.is_valid("AB:CD:EF:12:34:56")
    assert v.is_valid("ABCD:EF12:3456")
    assert v.is_valid("AB-CD-EF-12-34-56")
    assert v.is_valid("ABCD-EF12-3456")

    assert not v.is_valid("qwertyuiop")
    assert not v.is_valid("qw-er-ty-12-34-56")
    assert not v.is_valid("ab:cd:ef:12:34:56:78")
    assert not v.is_valid("abcdefghijkl")
    assert not v.is_valid("1234567890ax")


def test_semver():
    """
    - https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
    - https://regex101.com/r/Ly7O1x/3/
    """
    v = val.SemVer()

    assert v.is_valid("0.0.4")
    assert v.is_valid("1.2.3")
    assert v.is_valid("10.20.30")
    assert v.is_valid("1.1.2-prerelease+meta")
    assert v.is_valid("1.1.2+meta")
    assert v.is_valid("1.1.2+meta-valid")
    assert v.is_valid("1.0.0-alpha")
    assert v.is_valid("1.0.0-beta")
    assert v.is_valid("1.0.0-alpha.beta")
    assert v.is_valid("1.0.0-alpha.beta.1")
    assert v.is_valid("1.0.0-alpha.1")
    assert v.is_valid("1.0.0-alpha0.valid")
    assert v.is_valid("1.0.0-alpha.0valid")
    assert v.is_valid("1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay")
    assert v.is_valid("1.0.0-rc.1+build.1")
    assert v.is_valid("2.0.0-rc.1+build.123")
    assert v.is_valid("1.2.3-beta")
    assert v.is_valid("10.2.3-DEV-SNAPSHOT")
    assert v.is_valid("1.2.3-SNAPSHOT-123")
    assert v.is_valid("1.0.0")
    assert v.is_valid("2.0.0")
    assert v.is_valid("1.1.7")
    assert v.is_valid("2.0.0+build.1848")
    assert v.is_valid("2.0.1-alpha.1227")
    assert v.is_valid("1.0.0-alpha+beta")
    assert v.is_valid("1.2.3----RC-SNAPSHOT.12.9.1--.12+788")
    assert v.is_valid("1.2.3----R-S.12.9.1--.12+meta")
    assert v.is_valid("1.2.3----RC-SNAPSHOT.12.9.1--.12")
    assert v.is_valid("1.0.0+0.build.1-rc.10000aaa-kk-0.1")
    assert v.is_valid("99999999999999999999999.999999999999999999.99999999999999999")
    assert v.is_valid("1.0.0-0A.is.legal")

    assert not v.is_valid("1")
    assert not v.is_valid("1.2")
    assert not v.is_valid("1.2.3-0123")
    assert not v.is_valid("1.2.3-0123.0123")
    assert not v.is_valid("1.1.2+.123")
    assert not v.is_valid("+invalid")
    assert not v.is_valid("-invalid")
    assert not v.is_valid("-invalid+invalid")
    assert not v.is_valid("-invalid.01")
    assert not v.is_valid("alpha")
    assert not v.is_valid("alpha.beta")
    assert not v.is_valid("alpha.beta.1")
    assert not v.is_valid("alpha.1")
    assert not v.is_valid("alpha+beta")
    assert not v.is_valid("alpha_beta")
    assert not v.is_valid("alpha.")
    assert not v.is_valid("alpha..")
    assert not v.is_valid("beta")
    assert not v.is_valid("1.0.0-alpha_beta")
    assert not v.is_valid("-alpha.")
    assert not v.is_valid("1.0.0-alpha..")
    assert not v.is_valid("1.0.0-alpha..1")
    assert not v.is_valid("1.0.0-alpha...1")
    assert not v.is_valid("1.0.0-alpha....1")
    assert not v.is_valid("1.0.0-alpha.....1")
    assert not v.is_valid("1.0.0-alpha......1")
    assert not v.is_valid("1.0.0-alpha.......1")
    assert not v.is_valid("01.1.1")
    assert not v.is_valid("1.01.1")
    assert not v.is_valid("1.1.01")
    assert not v.is_valid("1.2")
    assert not v.is_valid("1.2.3.DEV")
    assert not v.is_valid("1.2-SNAPSHOT")
    assert not v.is_valid("1.2.31.2.3----RC-SNAPSHOT.12.09.1--..12+788")
    assert not v.is_valid("1.2-RC-SNAPSHOT")
    assert not v.is_valid("-1.0.3-gamma+b7718")
    assert not v.is_valid("+justmeta")
    assert not v.is_valid("9.8.7+meta+meta")
    assert not v.is_valid("9.8.7-whatever+meta+meta")



import ast

from .. import validators as val

safe_globals = ("True", "False", "None")
safe_builtins = dict((f, __builtins__[f]) for f in safe_globals)


def _validate_expr(call_node, validators):
    # Validate that the expression uses a known, registered validator.
    try:
        func_name = call_node.func.id
    except AttributeError:
        raise SyntaxError("Schema expressions must be enclosed by a validator.")
    if func_name not in validators:
        raise SyntaxError("Not a registered validator: '%s'. " % func_name)
    # Validate that all args are constant literals, validator names,  or other call nodes
    arg_values = call_node.args + [kw.value for kw in call_node.keywords]
    for arg in arg_values:
        base_arg = arg.operand if isinstance(arg, ast.UnaryOp) else arg
        if isinstance(base_arg, ast.Constant):
            continue
        elif isinstance(base_arg, ast.Name) and base_arg.id in validators:
            continue
        elif isinstance(base_arg, ast.Call):
            _validate_expr(base_arg, validators)
        else:
            raise SyntaxError("Argument values must either be constant literals, or else " "reference other validators.")


def parse(validator_string, validators=None):
    validators = validators or val.DefaultValidators
    try:
        tree = ast.parse(validator_string, mode="eval")
        _validate_expr(tree.body, validators)
        # evaluate with access to a limited global scope only
        return eval(compile(tree, "<ast>", "eval"), {"__builtins__": safe_builtins}, validators)
    except (SyntaxError, NameError, TypeError) as e:
        raise SyntaxError("Invalid schema expression: '%s'. " % validator_string + str(e))



from __future__ import absolute_import
from io import StringIO


def _pyyaml(f):
    import yaml

    try:
        Loader = yaml.CSafeLoader
    except AttributeError:  # System does not have libyaml
        Loader = yaml.SafeLoader
    return list(yaml.load_all(f, Loader=Loader))


def _ruamel(f):
    from ruamel.yaml import YAML

    yaml = YAML(typ="safe")
    return list(yaml.load_all(f))


_parsers = {"pyyaml": _pyyaml, "ruamel": _ruamel}


def parse_yaml(path=None, parser="pyyaml", content=None):
    try:
        parse = _parsers[parser.lower()]
    except KeyError:
        raise NameError('Parser "' + parser + '" is not supported\nAvailable parsers are listed below:\nPyYAML\nruamel')
    if (path is None and content is None) or (path is not None and content is not None):
        raise TypeError("Pass either path= or content=, not both")
    if path is not None:
        with open(path) as f:
            return parse(f)
    else:
        return parse(StringIO(content))



import pytest
from pytest import raises
from .. import parse_yaml

parsers = ["pyyaml", "PyYAML", "ruamel"]


@pytest.mark.parametrize("parser", parsers)
def test_reader_error(parser):
    with raises(IOError):
        parse_yaml("wat", parser)



from pytest import raises

from .. import parser as par
from yamale.validators.validators import (
    Validator,
    String,
    Regex,
    Number,
    Integer,
    Boolean,
    List,
    Day,
    Timestamp,
    Ip,
    Mac,
)


def test_eval():
    assert eval("String()") == String()


def test_types():
    assert par.parse("String()") == String()
    assert par.parse("str()") == String()
    assert par.parse("regex()") == Regex()
    assert par.parse("num()") == Number()
    assert par.parse("int()") == Integer()
    assert par.parse("day()") == Day()
    assert par.parse("timestamp()") == Timestamp()
    assert par.parse("bool()") == Boolean()
    assert par.parse("list(str())") == List(String())
    assert par.parse("ip()") == Ip()
    assert par.parse("mac()") == Mac()


def test_custom_type():
    class my_validator(Validator):
        pass

    assert par.parse("custom()", {"custom": my_validator}) == my_validator()


def test_required():
    assert par.parse("str(required=True)").is_required
    assert par.parse("str(required=False)").is_optional


def test_syntax_error():
    with raises(SyntaxError):
        par.parse("eval()")



class DataPath(object):
    def __init__(self, *path):
        self._path = path

    def __add__(self, other):
        dp = DataPath()
        dp._path = self._path + other._path
        return dp

    def __str__(self):
        return ".".join(map(str, (self._path)))

    def __repr__(self):
        return "DataPath({})".format(repr(self._path))



#!/usr/bin/env python
from setuptools import setup, find_packages

readme = open("README.md", encoding="utf-8").read()
license = open("LICENSE", encoding="utf-8").read()
version = open("yamale/VERSION", encoding="utf-8").read().strip()

setup(
    name="yamale",
    version=version,
    url="https://github.com/23andMe/Yamale",
    author="Bo Lopker",
    author_email="dev-api@23andme.com",
    description="A schema and validator for YAML.",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pyyaml"],
    extras_require={"ruamel": ["ruamel.yaml"]},
    python_requires=">=3.8",
    entry_points={
        "console_scripts": ["yamale=yamale.command_line:main"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)



class YamaleError(ValueError):
    def __init__(self, results):
        super(YamaleError, self).__init__("\n".join([str(x) for x in list(filter(lambda x: not x.isValid(), results))]))
        self.message = self.args[0]
        self.results = results
